import os
import sys
import json
import logging
import click
from pathlib import Path
from future.moves import subprocess

from mldock.config_managers.cli import \
    CliConfigureManager
import mldock.api.predict as predict_request
from mldock.terminal import \
    ChoiceWithNumbers, style_dropdown, style_2_level_detail, \
        ProgressLogger, pretty_build_logs
from mldock.platform_helpers.mldock import utils as mldock_utils
from mldock.config_managers.container import MLDockConfigManager
from mldock.api.local import \
    search_mldock_containers, docker_build, train_model, deploy_model

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')
MLDOCK_CONFIG_NAME = 'mldock.json'

@click.group()
def local():
    """
    Commands for local development
    """
    pass

@click.command()
@click.option('--dir', help='Set the working directory for your mldock container.', required=True)
@click.option('--no-cache', help='builds container from scratch', is_flag=True)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--stage', help='environment to stage.')
@click.pass_obj
def build(obj, dir, no_cache, tag, stage):
    """Command to build container locally
    """
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(dir, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_dir name
    mldock_config = mldock_manager.get_config()
    image_name = mldock_config.get("image_name", None)
    template_name = mldock_config.get("template", None)
    container_dir = mldock_config.get("container_dir", None)
    module_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
    )
    dockerfile_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
        container_dir
    )
    requirements_file_path = os.path.join(
        dir,
        mldock_config.get("requirements_dir", "requirements.txt")
    )

    with ProgressLogger(
        group='Stages',
        text='Retrieving Stages',
        spinner='dots',
        on_success='Stages Retrieved'
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]['tag']

        if tag is None:
            raise Exception("tag is not valid. Either choose a stage or set a tag manually")

    try:
        with ProgressLogger(group='Build', text='Building', spinner='dots') as spinner:
            logs = docker_build(
                image_name=image_name,
                dockerfile_path=dockerfile_path,
                module_path=module_path,
                target_dir_name=mldock_config.get("mldock_module_dir", "src"),
                requirements_file_path=requirements_file_path,
                no_cache=no_cache,
                docker_tag=tag,
                container_platform=template_name
            )
            for line in logs:

                spinner.info(pretty_build_logs(line=line))

                spinner.start()

    except subprocess.CalledProcessError as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

@click.command()
@click.option('--payload', default=None, help='path to payload file', required=True)
@click.option('--content-type', default='json', help='format of payload', type=click.Choice(['json', 'csv'], case_sensitive=False))
@click.option('--host', help='host url at which model is served', type=str, default='http://127.0.0.1:8080/invocations')
def predict(payload, content_type, host):
    """
    Command to execute prediction request against ml endpoint
    """
    with ProgressLogger(group='Predict', text='Running Request', spinner='dots') as spinner:
        if payload is None:
            logger.info("\nPayload cannot be None. Please provide path to payload file.")
        else:
            if content_type in ['application/json', 'json']:
                pretty_output = predict_request.send_json(filepath=payload, host=host)
            elif content_type in ['text/csv', 'csv']:
                pretty_output = predict_request.send_csv(filepath=payload, host=host)
            else:
                raise Exception("Content-type is not supported.")
            logger.info(pretty_output)

@click.command()
@click.option('--dir', help='Set the working directory for your mldock container.', required=True)
@click.option(
    '--params',
    '-p',
    help='(Optional) Hyperparameter override when running container.',
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True
)
@click.option(
    '--env_vars',
    '-e',
    help='(Optional) Hyperparameter override when running container.',
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True
)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--stage', help='environment to stage.')
@click.pass_obj
def train(obj, dir, params, env_vars, tag, stage):
    """
    Command to run training locally on localhost
    """
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(dir, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_path name
    mldock_config = mldock_manager.get_config()
    module_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
    )
    image_name = mldock_config.get("image_name", None)
    container_dir = mldock_config.get("container_dir", None)

    with ProgressLogger(
        group='Stages',
        text='Retrieving Stages',
        spinner='dots',
        on_success='Stages Retrieved'
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]['tag']

        if tag is None:
            raise Exception("tag is not valid. Either choose a stage or set a tag manually")

    with ProgressLogger(
        group='Environment',
        text='Retrieving Environment Varaiables',
        spinner='dots',
        on_success='Environment Ready'
    ) as spinner:
        project_env_vars = mldock_config.get('environment', {})

        for env_var in env_vars:
            key_, value_ = env_var
            project_env_vars.update(
                {key_: value_}
            )

        hyperparameters = mldock_config.get('hyperparameters', {})

        # override hyperparameters
        for param in params:
            key_, value_ = param
            hyperparameters.update(
                {key_: value_}
            )

        env_vars = mldock_utils.collect_mldock_environment_variables(
            stage=stage,
            hyperparameters=hyperparameters,
            **project_env_vars
        )

        config_manager = CliConfigureManager()
        env_vars.update(
            config_manager.local.get('environment')
        )

    with ProgressLogger(
        group='Training',
        text='Running Training',
        spinner='dots'
    ) as spinner:

        spinner.info("Training Environment = {}".format(env_vars))
        spinner.start()

        train_model(
            working_dir=dir,
            docker_tag=tag,
            image_name=image_name,
            entrypoint="src/container/executor.sh",
            cmd="train",
            env=env_vars
        )


@click.command()
@click.option('--dir', help='Set the working directory for your mldock container.', required=True)
@click.option(
    '--params',
    '-p',
    help='(Optional) Hyperparameter override when running container.',
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True
)
@click.option(
    '--env_vars',
    '-e',
    help='(Optional) Hyperparameter override when running container.',
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True
)
@click.option('--tag', help='docker tag', type=str, default='latest')
@click.option('--port', help='host url at which model is served', type=str, default='8080')
@click.option('--stage', help='environment to stage.')
@click.pass_obj
def deploy(obj, dir, params, env_vars, tag, port, stage):
    """
    Command to deploy ml container on localhost
    """
    verbose = obj.get('verbose', False)
    mldock_manager = MLDockConfigManager(
        filepath=os.path.join(dir, MLDOCK_CONFIG_NAME)
    )
    # get mldock_module_path name
    mldock_config = mldock_manager.get_config()
    module_path = os.path.join(
        dir,
        mldock_config.get("mldock_module_dir", "src"),
    )
    image_name = mldock_config.get("image_name", None)
    container_dir = mldock_config.get("container_dir", None)

    with ProgressLogger(
        group='Stages',
        text='Retrieving Stages',
        spinner='dots',
        on_success='Stages Retrieved'
    ) as spinner:
        stages = mldock_config.get("stages", None)

        if stage is not None:
            tag = stages[stage]['tag']

        if tag is None:
            raise Exception("tag is not valid. Either choose a stage or set a tag manually")

    project_env_vars = mldock_config.get('environment', {})

    for env_var in env_vars:
        key_, value_ = env_var
        print(key_, value_)
        project_env_vars.update(
            {key_: value_}
        )  
    hyperparameters = mldock_config.get('hyperparameters', {})

    # override hyperparameters
    for param in params:
        key_, value_ = param
        print(key_, value_)
        hyperparameters.update(
            {key_: value_}
        )

    env_vars = mldock_utils.collect_mldock_environment_variables(
        stage=stage,
        hyperparameters=mldock_config.get('hyperparameters', {}),
        **project_env_vars
    )

    config_manager = CliConfigureManager()
    env_vars.update(
        config_manager.local.get('environment')
    )

    with ProgressLogger(
        group='Deploy',
        text='Deploying model to {} @ localhost'.format(port),
        spinner='dots'
    ) as spinner:

        spinner.info("Deployment Environment = {}".format(env_vars))
        spinner.start()
        deploy_model(
            working_dir=dir,
            docker_tag=tag,
            image_name=image_name,
            port=port,
            entrypoint="src/container/executor.sh",
            cmd="serve",
            env=env_vars,
            verbose=verbose
        )

@click.command()
def stop():
    """
    Command to stop ml container on localhost
    """

    try:
        containers = search_mldock_containers()

        tags = [
            style_2_level_detail(
                major_detail=c.image.tags[0],
                minor_detail=c.name
             ) for c in containers
        ] + ['abort']
        # newline break
        click.echo("")
        click.secho("Running MLDock containers:", bg='blue', nl=True)
        container_tag = click.prompt(
            text=style_dropdown(group_name="container name", options=tags, default='abort'),
            type=ChoiceWithNumbers(tags, case_sensitive=False),
            show_default=False,
            default='abort',
            show_choices=False
        )
        project_container = None
        for c in containers:
            for image_tag in c.image.tags:
                if image_tag in container_tag:
                    # make newline
                    click.echo("")
                    project_container_tag = click.style(container_tag, fg='bright_blue')
                    if click.confirm(
                        "Stop the running container = {}?".format(
                            project_container_tag) 
                    ):
                        c.kill()

                    break

    except ValueError:
        logger.error("This is not a mldock directory: {}".format(dir))
        raise
    except subprocess.CalledProcessError as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

local.add_command(build)
local.add_command(predict)
local.add_command(train)
local.add_command(deploy)
local.add_command(stop)



"""
    DATAFRAME CONTENT DECODERS

    Handle Decoding of Content in to appropriate format from request as pandas dataframe.

    e.g. 
        list of lists -> pd.DataFrame.
"""
import io
import json
import pandas as pd

from mldock.platform_helpers.mldock.inference import content_types

def csv_to_pandas(bytes_like: bytes):
    """
        Decodes byte encoded csv to pandas dataframe.

        args:
            bytes_like (str): bytes serialized CSV string.
        return:
            (pd.DataFrame): data as pandas dataframe
    """
    s = bytes_like.decode()
    data = io.StringIO(s)
    return pd.read_csv(data)

def json_list_to_pandas(json_data):
    """
        Transforms a array of flat json objects in to pandas dataframe.

        args:
            json_data (str): data as json serialized list of dictionaries.
        return:
            (pd.DataFrame): data as pandas dataframe
    """
    data = json.loads(json_data)
    return pd.DataFrame(data)





"""Predict Request API utilties"""
import requests
import json

def send_json(filepath, host):
    """Send a json payload in request providing a filepath."""
    with open(filepath, 'r') as file_:
        data = json.load(file_)

    response = requests.post(
        host,
        headers={'Content-Type': 'application/json'},
        data=json.dumps(data)
    )
    if response.status_code == 200:
        return "{}".format(json.dumps(response.json(), indent=4, separators=(',', ': '), sort_keys=True))
    else:
        raise requests.exceptions.RequestException("error ({}): {}".format(response.status_code, response.raise_for_status()))

def send_csv(filepath, host):
    """Send a json payload in request providing a filepath."""
    with open(filepath, 'r') as file_:
        data = file_.read()

        response = requests.post(
            host,
            headers={'content-type': 'text/csv'},
            data=data
        )

    if response.status_code == 200:
        return response.text
    else:
        raise requests.exceptions.RequestException("error ({}): {}".format(response.status_code, response.raise_for_status()))


import os
import sys
import json
import logging
import click
from pathlib import Path
from future.moves import subprocess

from mldock.config_managers.core import WorkingDirectoryManager
from mldock.config_managers.container import MLDockConfigManager

from mldock.config_managers.cli import \
    ResourceConfigManager, PackageConfigManager, \
        HyperparameterConfigManager, InputDataConfigManager, StageConfigManager, \
            ModelConfigManager, EnvironmentConfigManager, CliConfigureManager

from mldock.platform_helpers import utils
from mldock.platform_helpers.mldock import utils as mldock_utils
from mldock.api.templates import init_from_template

click.disable_unicode_literals_warning = True
logger=logging.getLogger('mldock')
MLDOCK_CONFIG_NAME='mldock.json'


def reset_terminal():
    # os.system("clear")
    click.clear()

@click.group()
def container():
    """
    Commands to create, update and manage container projects and templates.
    """
    pass

@click.command()
@click.option('--dir', help='Set the working directory for your sagify container.', required=True)
@click.option('--no-prompt', is_flag=True, help='Do not prompt user, instead use the mldock config to initialize the container.')
@click.option('--container-only', is_flag=True, help='Only inject new container assets.')
@click.option('--template', default=None, help='Directory containing mldock supported container to use to initialize the container.')
@click.pass_obj
def init(obj, dir, no_prompt, container_only, template):
    """
    Command to initialize mldock enabled container project
    """
    reset_terminal()
    mldock_package_path = obj['mldock_package_path']
    try:
        click.secho("Initializing MLDock project configuration", bg='blue', nl=True)

        if not Path(dir).is_dir():
            create_new = click.prompt('No MLDOCK project found. Create?', type=bool)
        else:
            create_new = False

        mldock_manager = MLDockConfigManager(
            filepath=os.path.join(dir, MLDOCK_CONFIG_NAME),
            create=create_new
        )

        if template is not None:
            mldock_manager.update_config(template=template)

        if not no_prompt:
            mldock_manager.setup_config()

        package_manager = PackageConfigManager(
            filepath=os.path.join(dir, "requirements.txt"),
            create=True
        )
        # write to package manager
        package_manager.write_file()


        path_to_payload = Path(os.path.join(dir, "payload.json"))
        if not path_to_payload.exists():
            path_to_payload.write_text(json.dumps({"feature1": 10, "feature2":"groupA"}))

        # get sagify_module_path name
        mldock_config = mldock_manager.get_config()

        src_directory = os.path.join(
            dir,
            mldock_config.get("mldock_module_dir", "src")
        )
        mldock_module_path = os.path.join(
            src_directory,
            'container'
        )

        working_directory_manager = WorkingDirectoryManager(base_dir=dir)
        # sagemaker
        config_path = working_directory_manager.input_config_dir

        # create stages config
        stage_config_manager = StageConfigManager(
            config=mldock_config.get('stages', {})
        )
        # set input data channels
        input_data_channels = InputDataConfigManager(
            config=mldock_config.get('data', []),
            base_path=Path(dir, 'data')
        )

        # set model channels
        model_channels = ModelConfigManager(
            config=mldock_config.get('model', []),
            base_path=Path(dir, 'model')
        )

        # set hyperparameters
        hyperparameters = HyperparameterConfigManager(
            config=mldock_config.get('hyperparameters', {})
        )

        # set environment variables
        environment = EnvironmentConfigManager(
            config=mldock_config.get('environment', {})
        )

        if not no_prompt:
            stage_config_manager.ask_for_stages()
            # update stages
            mldock_manager.update_stages(stages=stage_config_manager.get_config())
            # update input data channels
            input_data_channels.ask_for_input_data_channels()
            input_data_channels.write_gitignore()
            mldock_manager.update_data_channels(data=input_data_channels.get_config())
            # update model channels
            model_channels.ask_for_model_channels()
            model_channels.write_gitignore()
            mldock_manager.update_model_channels(models=model_channels.get_config())
            # update hyperparameters
            hyperparameters.ask_for_hyperparameters()
            mldock_manager.update_hyperparameters(hyperparameters=hyperparameters.get_config())
            # update environments
            environment.ask_for_env_vars()
            mldock_manager.update_env_vars(environment=environment.get_config())

        mldock_manager.write_file()

        # Get template specific files
        template = mldock_config["template"] if template is None else templates

        config_manager = CliConfigureManager()
        templates = config_manager.templates

        # get configured template server metadata
        # if not set, default to package default templates
        templates_root = templates.get('templates_root', Path(mldock_package_path, 'templates'))
        # if not set, default to local
        template_server = templates.get('server_type', 'local')

        init_from_template(
            template_name=template,
            templates_root=templates_root,
            src_directory=src_directory,
            container_only=container_only,
            template_server=template_server
        )

        # reset_terminal()
        # logger.info(obj["logo"])
        states = mldock_manager.get_state()

        for state in states:
            click.echo(click.style(state["name"], bg='blue'), nl=True)
            click.echo(click.style(state["message"], fg='white'), nl=True)

        logger.info("\nlocal container volume is ready! ヽ(´▽`)/")
    except subprocess.CalledProcessError as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

@click.command()
@click.option('--dir', help='Set the working directory for your sagify container.', required=True)
@click.pass_obj
def update(obj, dir):
    """
    Command to update mldock container.
    """
    reset_terminal()
    try:
        logger.info("Loading MLDock config")
        mldock_manager = MLDockConfigManager(
            filepath=os.path.join(dir, MLDOCK_CONFIG_NAME)
        )

        # get sagify_module_path name
        mldock_config = mldock_manager.get_config()

        # get list of dataset names
        mldock_data = mldock_config.get("data", None)
        input_config_config = mldock_utils._extract_data_channels_from_mldock(mldock_data)

        src_directory = os.path.join(
            dir,
            mldock_config.get("mldock_module_dir", "src")
        )
        mldock_module_path = os.path.join(
            src_directory,
            'container'
        )

        working_directory_manager = WorkingDirectoryManager(base_dir=dir)
        # sagemaker
        config_path = working_directory_manager.input_config_dir

        # create stages config
        stage_config_manager = StageConfigManager(
            config=mldock_config.get('stages', {})
        )
        # set input data channels
        input_data_channels = InputDataConfigManager(
            config=mldock_config.get('data', []),
            base_path=Path(dir, 'data')
        )
        # set model channels
        model_channels = ModelConfigManager(
            config=mldock_config.get('model', []),
            base_path=Path(dir, 'model')
        )

        # set hyperparameters
        hyperparameters = HyperparameterConfigManager(
            config=mldock_config.get('hyperparameters', {})
        )

        # set environment variables
        environment = EnvironmentConfigManager(
            config=mldock_config.get('environment', {})
        )

        stage_config_manager.ask_for_stages()
        # update stages
        mldock_manager.update_stages(stages=stage_config_manager.get_config())
        # update input data channels
        input_data_channels.ask_for_input_data_channels()
        input_data_channels.write_gitignore()
        mldock_manager.update_data_channels(data=input_data_channels.get_config())
        # update model channels
        model_channels.ask_for_model_channels()
        model_channels.write_gitignore()
        mldock_manager.update_model_channels(models=model_channels.get_config())
        # update hyperparameters
        hyperparameters.ask_for_hyperparameters()
        mldock_manager.update_hyperparameters(hyperparameters=hyperparameters.get_config())
        # update environments
        environment.ask_for_env_vars()
        mldock_manager.update_env_vars(environment=environment.get_config())

        mldock_manager.write_file()

        click.clear()
        logger.info(obj["logo"])
        states = mldock_manager.get_state()

        for state in states:
            click.echo(click.style(state["name"], bg='blue'), nl=True)
            click.echo(click.style(state["message"], fg='white'), nl=True)

        logger.info("\nlocal container was updated! ヽ(´▽`)/")
    except subprocess.CalledProcessError as exception:
        logger.error(exception)
        raise
    except Exception as exception:
        logger.error(exception)
        raise

@click.command()
@click.option('--dir', help='Set the working directory for your sagify container.', required=True)
@click.pass_obj
def summary(obj, dir):
    """
    Command to show summary for mldock container
    """
    try:
        logger.info("Loading MLDock config")
        mldock_manager = MLDockConfigManager(
            filepath=os.path.join(dir, MLDOCK_CONFIG_NAME)
        )

        states = mldock_manager.get_state()

        for state in states:
            click.echo(click.style(state["name"], bg='blue'), nl=True)
            click.echo(click.style(state["message"], fg='white'), nl=True)

        logger.info("\nlocal container was updated! ヽ(´▽`)/")
    except Exception as exception:
        logger.error(exception)
        raise

def add_commands(cli):
    """add cli commands to group"""
    cli.add_command(init)
    cli.add_command(update)
    cli.add_command(summary)

add_commands(container)



import os
import sys
import time
from click.testing import CliRunner
import pytest
from serum.__main__ import cli
from serum.api.local import local_deploy_detached

class TestContainerHealth:

    project_dir=/add/you/project/dir/here

    cli_dir='--dir={PROJECT_DIR}'.format(PROJECT_DIR=project_dir)

    def test_container_build_success(self, capsys):
        """Test container builds successfull
        """
        runner = CliRunner()

        result = runner.invoke(cli=cli, args=['local', 'build', self.cli_dir])

        assert result.exit_code == 0
        assert "Build Complete! ヽ(´▽`)/" in result.output

    def test_container_run_train_success(self, capsys):
        """Test container runs training successfully
        """
        runner = CliRunner()

        result = runner.invoke(cli=cli, args=['local', 'train', self.cli_dir])

        assert result.exit_code == 0
        assert "Training Complete! ヽ(´▽`)/" in result.output

    def test_container_run_deploy_success(self, capsys):
        """Test container runs deploy successfully
        """
        runner = CliRunner()

        result = runner.invoke(cli=cli, args=['local', 'deploy', self.cli_dir])

        assert result.exit_code == 0
        assert "Deploy Ready! ヽ(´▽`)/" in result.output




class Credentials(object):

    @staticmethod
    def refresh(request, **kwargs):

        pass

    @property
    def token(self):
        return 'PASSWORD'

def default(scopes: list, **kwargs):
    return Credentials(), 'myproject'

class Request(object):
    pass



"""
    Docker Auth Utilities aand workflows
"""
import os
import logging
from urllib.parse import urlparse
import docker

from mldock.platform_helpers.aws.auth import get_aws_ecr
from mldock.platform_helpers.gcp.auth import get_gcp_gcr
from mldock.platform_helpers.dockerhub.auth import get_dockerhub_credentials

logger = logging.getLogger('mldock')

def login_and_authenticate(provider: str, region: str):
    """
    Login and authenticate with registry

    DockerHub - https://ropenscilabs.github.io/r-docker-tutorial/04-Dockerhub.html
    AWS - https://aws.amazon.com/blogs/compute/authenticating-amazon-ecr-repositories-for-docker-cli-with-credential-helper/
    GCP - https://cloud.google.com/artifact-registry/docs/docker/authentication
    Azure - https://docs.microsoft.com/en-us/azure/container-registry/container-registry-repository-scoped-permissions
    """
    client = docker.from_env()

    if provider == 'ecr':
        # AWS ECR
        username, password, registry, repo = get_aws_ecr(region=region)
    elif provider == 'gcr':
        # GCP 
        username, password, registry, repo = get_gcp_gcr(region=region)
    elif provider == 'dockerhub':
        # DockerHub
        username, password, registry, repo = get_dockerhub_credentials()
    else:
        raise TypeError("{PROVIDER} Registry is not implemented.".format(PROVIDER=provider))

    # login
    response = client.login(
        username=username,
        password=password,
        email=None,
        registry=registry,
        reauth=True
    )

    # return logged in client
    return client, {
        'username': username,
        'password': password,
        'registry': registry,
        'repository': repo
    }



"""
    Config managers for Configure commands

    Config managers used in command/configure
"""
import os
import json
import yaml
import click
import logging
from pathlib import Path
from appdirs import user_config_dir
from mldock.terminal import ChoiceWithNumbers, style_dropdown
from mldock.config_managers.core import BaseConfigManager

APP_NAME = 'mldock'
APP_AUTHOR = 'mldock'

logger = logging.getLogger('mldock')

class CliConfigureManager(BaseConfigManager):
    """CLI Configure Manager for mldock
    """

    config = {}

    def __init__(
        self,
        filepath: str = user_config_dir(
            appname=APP_NAME, appauthor=APP_AUTHOR
        ),
        create: bool = False
    ):
        self.filepath = filepath
        self.config = self.load_config(self.filepath, create=create)

    def reset(self, config_type):
        click.secho("Dropping {} configuration".format(config_type), bg='blue')
        self.config.pop(config_type, None)

    def setup_local_config(self):
        click.secho("Setup local CLI configuration", bg='blue')
        if self.config.get('local') is None:
            self.config['local'] = {}
        self.ask_for_container_auth_type()
        self.ask_for_environment(config_type='local')

    def setup_templates_config(self):
        click.secho("Setup templates CLI configuration", bg='blue')
        if self.config.get('templates') is None:
            self.config['templates'] = {}
        self.ask_for_template_server()
        self.ask_for_template_root()

    def setup_workspace_config(self):
        click.secho("Setup workspace CLI configuration", bg='blue')
        if self.config.get('workspace') is None:
            self.config['workspace'] = {}
        self.ask_for_environment(config_type='workspace')

    def ask_for_container_auth_type(self):
        """prompt user for platform name
        """

        click.secho("Set local authentication type", bg='blue', nl=True)
        options = ['gcloud', 'awscli']

        auth_type = click.prompt(
            text=style_dropdown(
                group_name="authentication type",
                options=options,
                default=self.config.get('auth_type', None)
            ),
            type=ChoiceWithNumbers(options, case_sensitive=False),
            show_default=True,
            default=self.config.get('auth_type', None),
            show_choices=False
        )
        
        self.config['local'].update({
            'auth_type': auth_type
        })

    def ask_for_environment(self, config_type):
        
        config = self.config[config_type].get('environment', {})
        environment_config_manager = EnvironmentConfigManager(config=config)
        environment_config_manager.ask_for_env_vars()
        self.config[config_type].update({
            'environment': environment_config_manager.get_config()
        })
    
    def ask_for_template_server(self):
        """prompt user to set template server tracking
        """

        click.secho("Set template server type", bg='blue', nl=True)
        options = ['local', 'github']

        server_type = click.prompt(
            text=style_dropdown(
                group_name="template server type",
                options=options,
                default=self.config.get('server_type', None)
            ),
            type=ChoiceWithNumbers(options, case_sensitive=False),
            show_default=True,
            default=self.config.get('server_type', None),
            show_choices=False
        )
        
        self.config['templates'].update({
            'server_type': server_type
        })

    def ask_for_template_root(self):
        """prompt user to set template root
        """

        click.secho("Set template root", bg='blue', nl=True)

        templates_root_dir = click.prompt(
            text=click.style("Set path to template root dir: ", fg='bright_blue'),
            default=self.config['templates'].get('templates_root', None)
        )
        
        self.config['templates'].update({
            'templates_root': templates_root_dir
        })

    @staticmethod
    def _format_nodes(config):

        output = []

        for key_ in sorted(config):
            value_ = config[key_]
            if isinstance(value_, dict):
                value_ = yaml.dump(value_, indent=4, sort_keys=True)
            output.append({
                "name": key_, "message": value_
            })

        return output

    def get_state(self):
        """pretty prints a json config to terminal
        """
        config = self.config.copy()

        states = self._format_nodes(config)

        return states

    @property
    def local(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get('local', {})

    @property
    def workspace(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get('workspace', {})

    @property
    def templates(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config.get('templates', {})

class ResourceConfigManager(BaseConfigManager):
    """Resource Config Manager for mldock
    """
    config = {
        "current_host": "algo-1",
        "hosts": [
            "algo-1"
        ],
        "network_interface_name": "eth1"
    }

    @staticmethod
    def ask_for_current_host_name():
        """prompt user for current host name

        Returns:
            return: current host name
        """
        current_host_name = chosen_python_index = click.prompt(
            text="Set current host name: ",
            default="algo",
        )

        return current_host_name

    @staticmethod
    def ask_for_network_interface_name():
        """prompt user for network interface name

        Returns:
            str: network interface name
        """
        network_interface_name = click.prompt(
            text="Set current host name: ",
            default="eth1",
        )

        return network_interface_name

    def ask_for_resourceconfig(self):
        """prompt user for resource config
        """
        current_host_name = self.ask_for_current_host_name()
        hosts = [current_host_name+"-1"]
        network_interface_name = self.ask_for_network_interface_name()

        self.config = {
            "current_host": hosts[0],
            "hosts": hosts,
            "network_interface_name": network_interface_name
        }

class PackageConfigManager(BaseConfigManager):
    """Package Requirement Config Manager for sagify
    """
    @staticmethod
    def touch(filename):
        """creat an empty txt file

        Args:
            filename (str): path to file
        """
        Path(filename).touch()

    def load_config(self, file_name: str, create: bool) -> list:
        """load config

        Args:
            filename (str): path to config file to load

        Returns:
            list: config
        """
        self.check_if_exists_else_create(
            file_name=file_name, create=create
        )
        path = Path(file_name)
        with path.open() as f: 
            config = f.readlines()
        
        for index, c_package in enumerate(config):
            if c_package.endswith("\n"):
                config[index] = config[index].split("\n")[0]
        return config

    def get_config(self) -> list:
        """get config object

        Returns:
            list: config
        """
        return self.config

    def seed_packages(self, packages: list):
        """seeds config with required package modules

        Args:
            packages (list): packages to add to config
        """
        for new_package in packages:
            for config_package in self.config:
                if new_package not in config_package:
                    self.config.append(new_package)

    def pretty_print(self):
        """pretty prints a list config to terminal
        """
        pretty_config = json.dumps(self.config, indent=4, separators=(',', ': '))
        logger.debug("{}\n".format(pretty_config))

    def write_file(self):
        """
        write to file
        """
        config = set(self.config)
        config_txt = "\n".join(config) + "\n"
        Path(self.filepath).write_text(config_txt)

class StageConfigManager(BaseConfigManager):
    """Development Stage Config Manager for mldock
    """

    def __init__(self, config: dict):
        self.config = config

    def ask_for_stages(self):
        """prompt user for stages
        """
        click.secho("Stages", bg='blue', nl=True)
        while True:
            click.echo(click.style("Add a development stage. ", fg='bright_blue')+"(Follow the prompts)", nl=True)
            stage_name = click.prompt(
                text="Stage name: ",
                default="end",
                show_default=False,
                type=str
            )
            if stage_name == "end":
                break
            else:
                docker_tag = click.prompt(
                    text="Set docker image tag: ",
                    default=self.config.get(stage_name, {}).get('tag', None)
                )
                if stage_name not in self.config:
                    self.config[stage_name] = {}

                self.config[stage_name].update(
                    {'tag': docker_tag}
                )

class HyperparameterConfigManager(BaseConfigManager):
    """Hyperparameter Config Manager for mldock
    """

    def __init__(self, config: dict):
        self.config = config

    @staticmethod
    def is_float(s):
        """ Returns True is string is a number. """
        try:
            float(s)
            return True
        except ValueError:
            return False

    def ask_for_hyperparameters(self):
        """prompt user for hyperparameters
        """
        click.secho("Hyperparameters", bg='blue', nl=True)
        while True:
            click.echo(click.style("Add a hyperparameter. ", fg='bright_blue')+"(Follow prompts)", nl=True)
            hparam_name = click.prompt(
                text="Hyperparameter (name): ",
                default="end",
                show_default=False,
                type=str
            )
            if hparam_name == "end":
                break
            else:
                hparam_value = click.prompt(
                    text="Set value: ",
                    default=self.config.get(hparam_name, None)
                )

                self.config.update(
                    {hparam_name: hparam_value}
                )

class EnvironmentConfigManager(BaseConfigManager):
    """Environment Config Manager for mldock
    """
    def __init__(self, config: dict):
        self.config = config

    @staticmethod
    def is_float(s):
        """ Returns True is string is a number. """
        try:
            float(s)
            return True
        except ValueError:
            return False

    @staticmethod
    def parse_bool(s):
        """ Returns True is string is a bool. """
        if s.lower() == 'true':
            return True
        elif s.lower() == 'false':
            return False
        else:
            return s

    def ask_for_env_vars(self):
        """prompt user for environment variables
        """
        click.secho("Environment Variables", bg='blue', nl=True)
        while True:
            click.echo(click.style("Add a environment variable. ", fg='bright_blue')+"(Follow prompts)", nl=True)
            env_var_name = click.prompt(
                text="Environment Variable (name): ",
                default="end",
                show_default=False,
                type=str
            )
            if env_var_name == "end":
                break
            else:
                env_var_value = click.prompt(
                    text="Set value: ",
                    default=self.config.get(env_var_name, None)
                )

                self.config.update(
                    {env_var_name: env_var_value}
                )

class InputDataConfigManager(BaseConfigManager):
    """InputData Config Manager for mldock
    """
    file_ignores = []
    def __init__(self, config: dict, base_path: str):
        self.config = config
        self.base_path = base_path

    def write_gitignore(self):
        """
        write to file
        """
        config = set([Path(dataset['channel'], dataset['filename']).as_posix() for dataset in self.config])
        config_txt = "\n".join(config) + "\n"

        ignore_filepath = Path(self.base_path,".gitignore")
        Path(ignore_filepath).write_text(config_txt)

    def ask_for_input_data_channels(self):
        """prompt user for hyperparameters
        """
        click.secho("Input Data Channels", bg='blue', nl=True)
        while True:
            channel_filename_pair = click.prompt(
                text=click.style("Add a data channel. ", fg='bright_blue')+"(Expects channel/filename). Hit enter to continue.",
                default="end",
                show_default=False,
                type=str
            )
            if channel_filename_pair == "end":
                logger.debug("\nUpdated data channels")
                self.pretty_print()
                break
            elif "/" in channel_filename_pair and not channel_filename_pair == "channel:filename":
                channel, filename = channel_filename_pair.split("/",1)
                logger.debug("Adding data/{}/{}".format(channel, filename))
                self.config.append({
                    'channel': channel,
                    'filename': filename
                })
            else:
                logger.warning("Expected format as channel/filename. Skipping")

class ModelConfigManager(BaseConfigManager):
    """Model Config Manager for mldock
    """
    file_ignores = []
    def __init__(self, config: dict, base_path: str):
        self.config = config
        self.base_path = base_path

    def write_gitignore(self):
        """
        write to file
        """
        config = set([Path(dataset['channel'], dataset['filename']).as_posix() for dataset in self.config])
        config_txt = "\n".join(config) + "\n"

        ignore_filepath = Path(self.base_path,".gitignore")
        Path(ignore_filepath).write_text(config_txt)

    def ask_for_model_channels(self):
        """prompt user for hyperparameters
        """
        click.secho("Model Channels", bg='blue', nl=True)
        while True:
            channel_filename_pair = click.prompt(
                text=click.style("Add a model channel. ", fg='bright_blue')+"(Expects channel/filename). Hit enter to continue.",
                default="end",
                show_default=False,
                type=str
            )
            if channel_filename_pair == "end":
                logger.debug("\nUpdated model channels")
                self.pretty_print()
                break
            elif "/" in channel_filename_pair and not channel_filename_pair == "channel:filename":
                channel, filename = channel_filename_pair.split("/",1)
                logger.debug("Adding data/{}/{}".format(channel, filename))
                self.config.append({
                    'channel': channel,
                    'filename': filename
                })
            else:
                logger.warning("Expected format as channel/filename. Skipping")



"""Authenticate docker to dockerhub"""
from environs import Env

def get_dockerhub_credentials():
    """
        Get Authentication credentials for Dockerhub

        return:
            username (str): username to use in docker client auth
            password (str): password token to use in docker client auth
            registry (str): registry url to use
    """
    # pick up from environment otherwise raise error
    env = Env()
    env.read_env()

    username = env("DOCKERHUB_USERNAME")
    password = env("DOCKERHUB_PASSWORD")
    registry = env("DOCKERHUB_REGISTRY")
    repo = env("DOCKERHUB_REPO")
    # return docker credentials
    return username, password, registry, repo



"""
    GCP Auth utilities
"""
import os
from google.auth import default
from google.auth.transport.requests import Request

from mldock.platform_helpers.utils import strip_scheme

def get_gcp_gcr(region: str):
    """
        Get Authentication credentials for google.auth

        args:
            region (str): Region to Authenticate GCR registry in
        
        return:
            username (str): username to use in docker client auth
            password (str): password token to use in docker client auth
            registry (str): registry url to use
    """
    credentials, project = default(
        scopes=['https://www.googleapis.com/auth/cloud-platform']
    )
    # creds.valid is False, and creds.token is None
    # Need to refresh credentials to populate those
    auth_req = Request()
    credentials.refresh(auth_req)
    # set docker creds
    username = 'oauth2accesstoken'
    password = credentials.token
    if region == 'eu':
        registry = 'https://eu.gcr.io'
    elif region == 'us':
        registry = 'https://us.gcr.io'
    elif region == 'asia':
        registry = 'https://asia.gcr.io'
    else:
        registry = 'https://gcr.io'

    registry = os.path.join(registry, project)
    # return docker credentials
    cloud_repository = strip_scheme(registry)
    # return docker credentials
    return username, password, registry, cloud_repository



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License'). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the 'license' file accompanying this file. This file is
# distributed on an 'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
#
# Disclaimer: This code can be found here: https://github.com/aws/sagemaker-training-toolkit/blob/master/test/unit/test_encoder.py
#
import io
import itertools

from mock import Mock, patch
import numpy as np
import pytest
from scipy import sparse
from six import BytesIO

from mldock.platform_helpers.mldock.inference import content_types
from mldock.platform_helpers.mldock.inference.content_decoders import numpy as numpy_decoders
from mldock.platform_helpers.mldock.inference.record_pb2 import Record
from mldock.platform_helpers.mldock.inference.recordio import _read_recordio


class TestNumpyDecoders:
    """Tests the numpy decoder methods"""

    @staticmethod
    @pytest.mark.parametrize(
        "target",
        ([42, 6, 9], [42.0, 6.0, 9.0], ["42", "6", "9"], [u"42", u"6", u"9"], {42: {"6": 9.0}}),
    )
    def test_npy_to_numpy(target):
        buffer = BytesIO()
        np.save(buffer, target)
        input_data = buffer.getvalue()

        actual = numpy_decoders.npy_to_numpy(input_data)

        np.testing.assert_equal(actual, np.array(target))

    @staticmethod
    @pytest.mark.parametrize(
        "target, expected",
        [
            ("[42, 6, 9]", np.array([42, 6, 9])),
            ("[42.0, 6.0, 9.0]", np.array([42.0, 6.0, 9.0])),
            ('["42", "6", "9"]', np.array(["42", "6", "9"])),
            (u'["42", "6", "9"]', np.array([u"42", u"6", u"9"])),
        ],
    )
    def test_json_to_numpy(target, expected):
        """test json data is correctly decoded to numpy array"""
        actual = numpy_decoders.json_to_numpy(target)
        np.testing.assert_equal(actual, expected)

    @staticmethod
    @pytest.mark.parametrize(
        "target, expected",
        [
            (b"42\n6\n9\n", np.array([42, 6, 9]).astype(str)),
            (b"42.0\n6.0\n9.0\n", np.array([42.0, 6.0, 9.0]).astype(str)),
            (b"42\n6\n9\n", np.array([42, 6, 9]).astype(str)),
            (b'"False,"\n"True."\n"False,"\n', np.array(["False,", "True.", "False,"])),
            (b'aaa\n"b""bb"\nccc\n', np.array(["aaa", 'b"bb', "ccc"])),
            (b'"a\nb"\nc\n', np.array(["a\nb", "c"])),
        ],
    )
    def test_csv_to_numpy(target, expected):
        """test csv data is correctly decoded to numpy array"""
        actual = numpy_decoders.csv_to_numpy(target)

        np.testing.assert_equal(actual, expected)




# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import sys
import logging
import click

PY2 = sys.version_info[0] == 2

if PY2:
    text_type = unicode  # noqa
else:
    text_type = str

class ColorFormatter(logging.Formatter):
    colors = {
        'error': dict(fg='red'),
        'exception': dict(fg='red'),
        'critical': dict(fg='red'),
        'debug': dict(fg='blue'),
        'warning': dict(fg='yellow')
    }

    def format(self, record):
        if not record.exc_info:
            level = record.levelname.lower()
            msg = record.msg
            if level in self.colors:
                prefix = click.style('{}: '.format(level),
                                     **self.colors[level])
                if not PY2 and isinstance(msg, bytes):
                    msg = msg.decode(sys.getfilesystemencoding(),
                                     'replace')
                elif not isinstance(msg, (text_type, bytes)):
                    msg = str(msg)
                msg = '\n'.join(prefix + x for x in msg.splitlines())
            return msg
        return logging.Formatter.format(self, record)

class ClickHandler(logging.Handler):
    def emit(self, record):
        try:
            msg = self.format(record)
            level = record.levelname.lower()
            err = level in ('warning', 'error', 'exception', 'critical')
            click.echo(msg, err=err)
        except Exception:
            self.handleError(record)

def _normalize_logger(logger, log_level):
    if not isinstance(logger, logging.Logger):
        logger = logging.getLogger(logger)

    logger.setLevel(log_level)
    return logger

def configure_logger(logger, verbose=False):
    log_level = logging.DEBUG if verbose else logging.INFO

    logger = _normalize_logger(logger, log_level)
    _default_handler = ClickHandler()
    _default_handler.formatter = ColorFormatter()
    logger.handlers = [_default_handler]
    logger.propagate = False
    return logger



import os
from pathlib import Path
import logging
import click

from mldock.config_managers.cli import \
    CliConfigureManager

click.disable_unicode_literals_warning = True
logger=logging.getLogger('mldock')

def reset_terminal():

    click.clear()

@click.group()
def configure():
    """
    Commands to configure mldock cli.
    """
    pass

@click.command()
@click.option(
    '-c', '--config-name',
    help='Configuration name you wish to reset',
    required=True,
    type=click.Choice(
        ['local', 'workspace', 'templates', 'all'],
        case_sensitive=False
    )
)
@click.pass_obj
def reset(obj, config_name):
    """reset configurations"""
    
    config_manager = CliConfigureManager()
    if config_name == 'all':
        for config in ['local', 'workspace', 'templates']:
            config_manager.reset(config)
    else:
        config_manager.reset(config_name)
    config_manager.write_file()

    reset_terminal()
    logger.info(obj['logo'])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg='blue'), nl=True)
        click.echo(click.style(state["message"], fg='white'), nl=True)


@click.command()
@click.pass_obj
def workspace(obj):
    """Configure for local development tools"""
    
    config_manager = CliConfigureManager(create=True)
    config_manager.setup_workspace_config()
    config_manager.write_file()

    reset_terminal()
    logger.info(obj['logo'])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg='blue'), nl=True)
        click.echo(click.style(state["message"], fg='white'), nl=True)

@click.command()
@click.pass_obj
def local(obj):
    """Configure for local development tools"""
    
    config_manager = CliConfigureManager(create=True)
    config_manager.setup_local_config()
    config_manager.write_file()

    reset_terminal()
    logger.info(obj['logo'])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg='blue'), nl=True)
        click.echo(click.style(state["message"], fg='white'), nl=True)

@click.command()
@click.pass_obj
def templates(obj):
    """Configure for local development tools"""
    
    config_manager = CliConfigureManager(create=True)
    config_manager.setup_templates_config()
    config_manager.write_file()

    reset_terminal()
    logger.info(obj['logo'])
    states = config_manager.get_state()

    for state in states:
        click.echo(click.style(state["name"], bg='blue'), nl=True)
        click.echo(click.style(state["message"], fg='white'), nl=True)

@click.command()
@click.pass_obj
def show(obj):
    """Configure for local development tools"""
    try:
        config_manager = CliConfigureManager()

        states = config_manager.get_state()

        for state in states:
            click.echo(click.style(state["name"], bg='blue'), nl=True)
            click.echo(click.style(state["message"], fg='white'), nl=True)
    except FileNotFoundError as exception:
        if Path(exception.filename).name == 'mldock':
            logger.error("File not found. Please run 'mldock configure local' to generate cli configuration.")
        else:
            raise

configure.add_command(workspace)
configure.add_command(local)
configure.add_command(templates)
configure.add_command(show)
configure.add_command(reset)



import os
import environs
import json
from pathlib import Path
import logging

from mldock.platform_helpers import utils


logger = logging.getLogger('mldock')

class AbstractEnvironment:

    def get_input_channel_iter(self):
        """Returns: Iterable of the input channels
        """
        return list()

    def get_output_channel_iter(self):
        """Returns: Iterable of the output channel
        """
        return list()

    def get_model_input_channel_iter(self):
        """Returns: Iterable of the model input channels
        """
        return list()
        
    def get_model_output_channel_iter(self):
        """Returns: Iterable of the model output channels
        """
        return list()

    def setup_inputs(self):
        """ Iterates and downloads assets remoate -> input channels
            return:
                None
        """
        pass

    def cleanup_outputs(self):
        """ Iterates and uploads output channel -> remote
            return:
                None
        """
        pass

    def setup_model_artifacts(self):
        """ Iterates and downloads assets remoate -> model channel
            return:
                None
        """
        pass

    def cleanup_model_artifacts(self):
        """ Iterates and uploads from model channel -> remote
            return:
                None
        """
        pass


class BaseEnvironment(AbstractEnvironment):

    input_channel_regex = None
    model_channel_regex = None
    output_channel_regex = None
    output_channel_regex = None
    hyperparameters_file = None
    environment_variables = environs.Env()

    def __init__(self, base_dir, **kwargs):

        self.environment_variables.read_env()
        self.base_dir = base_dir
        self.hyperparameters_file = kwargs.get(
            'hyperparameters_file', 'hyperparameters.json'
        )

        self.input_channel_regex = kwargs.get(
            'input_channel_regex', r'MLDOCK_INPUT_CHANNEL_.*'
        )
        self.model_input_channel_regex = kwargs.get(
            'model_input_channel_regex', r'MLDOCK_MODEL_INPUT_CHANNEL.*'
        )

        self.model_output_channel_regex = kwargs.get(
            'model_output_channel_regex', r'MLDOCK_MODEL_OUTPUT_CHANNEL.*'
        )

        self.output_channel_regex = kwargs.get(
            'output_channel_regex', r'MLDOCK_OUTPUT_CHANNEL_.*'
        )
        self.hyperparamters_env_variable = kwargs.get(
            'hyperparamters_regex', r'MLDOCK_HYPERPARAMETERS'
        )

        self._create_training_directories()
        self.setup_hyperparameters()

    def _create_training_directories(self):
        """Create the directory structure, if not exists
        """
        logger.debug("Creating a new training folder under {} .".format(self.base_dir))

        try:
            self.input_data_dir.mkdir(parents=True, exist_ok=True)
            self.model_dir.mkdir(parents=True, exist_ok=True)
            self.input_config_dir.mkdir(parents=True, exist_ok=True)
            self.output_data_dir.mkdir(parents=True, exist_ok=True)
            if not self.hyperparameters_filepath.exists():
                utils._write_json(
                    obj={},
                    path=self.hyperparameters_filepath
                )
        except Exception as exception:
            logger.error(exception)

    @property
    def input_dir(self):
        return Path(self.base_dir, 'input')

    @property
    def input_data_dir(self):
        return Path(self.input_dir, 'data')

    @property
    def input_config_dir(self):
        return Path(self.input_dir, 'config')

    @property
    def hyperparameters_filepath(self):
        return Path(self.input_config_dir, self.hyperparameters_file)

    @property
    def model_dir(self):
        return Path(self.base_dir, 'model')
    
    @property
    def output_data_dir(self):
        return Path(self.base_dir, 'output')

    @property
    def hyperparameters(self):
        """Returns: Iterable of the input channels
        """
        return utils._read_json(
            self.hyperparameters_filepath
        )

    def setup_hyperparameters(self):
        """Retrieves the env vars matching hyperparameters regex and updates config"""
        hyperparameters = utils._read_json(
            self.hyperparameters_filepath.as_posix()
        )

        updated_hparams = self.environment_variables.json(self.hyperparamters_env_variable, '{}')

        hyperparameters.update(
            updated_hparams
        )

        utils._write_json(
            obj=hyperparameters,
            path=self.hyperparameters_filepath.as_posix()
        )

    def get_input_channel_iter(self):
        """Returns: Iterable of the input channels
        """
        envvars = dict(os.environ)
        return utils.get_env_vars(environment=envvars, regex=self.input_channel_regex)

    def get_output_channel_iter(self):
        """Returns: Iterable of the output channel
        """
        envvars = dict(os.environ)
        return utils.get_env_vars(environment=envvars, regex=self.output_channel_regex)

    def get_model_input_channel_iter(self):
        """Returns: Iterable of the model channel
        """
        envvars = dict(os.environ)
        return utils.get_env_vars(environment=envvars, regex=self.model_input_channel_regex)

    def get_model_output_channel_iter(self):
        """Returns: Iterable of the model channel
        """
        envvars = dict(os.environ)
        return utils.get_env_vars(environment=envvars, regex=self.model_output_channel_regex)


"""MLDock Friendly Environments for Sagemaker"""
from pathlib import Path
from mldock.platform_helpers.mldock.configuration.environment import base
from mldock.platform_helpers.aws.sagemaker.environment import Environment as BaseSagemakerEnvironment

class SagemakerEnvironment(BaseSagemakerEnvironment, base.AbstractEnvironment):
    """
        Structures Sagemaker Environment similarily to
        other environments maintaining it's function
    """
    def __init__(self, **kwargs):
        super(SagemakerEnvironment, self).__init__()

    @property
    def input_data_dir(self):
        """
            Input data directory where channels are stored.
        
            Note:
                - this extends the sagemaker API to the standard mldock API.
        """
        return Path(self.input_dir, 'data')



"""MLDock Friendly Environments for GCP"""
from pathlib import Path
import logging

from mldock.platform_helpers.mldock.configuration.environment import base
from mldock.platform_helpers.gcp.storage import \
    download_input_assets, package_and_upload_model_dir, package_and_upload_output_data_dir

logger = logging.getLogger('mldock')

class GCPEnvironment(base.BaseEnvironment):
    """
        Extends the Environment class to give us
        more specific environment configuration tasks based on
        the GCP resource eco-system.
    """

    def setup_inputs(self):
        """ Iterates and downloads assets remoate -> input channels
        """
        logger.info("Setup assets in {}".format(self.input_data_dir))
        # only fetch channels of environment prefix MLDOCK_INPUT_CHANNEL_
        channels = self.get_input_channel_iter()

        if len(channels) == 0:
            logger.info("No input channels were found in ENV VARS.")

        for channel in channels:

            try:
                download_input_assets(
                    storage_dir_path=channel['value'],
                    local_path=self.input_data_dir,
                    scheme='gs'
                )

            except FileExistsError as exception:
                logger.info("{} Channel skipped. Already exists".format(channel['key']))

    def cleanup_outputs(self):
        """ Iterates and uploads output channel -> remote
        """
        logger.info("Cleanup assets in {}".format(self.output_data_dir))
        # only fetch channels of environment prefix MLDOCK_OUTPUT_CHANNEL_
        channels = self.get_output_channel_iter()

        if len(channels) == 0:
            logger.info("No output channels were found in ENV VARS.")

        for channel in channels:
            # only fetch channels with output
            channel_path = channel['key'].replace("MLDOCK_OUTPUT_CHANNEL_", "").lower()
            local_channel_path = Path(self.output_data_dir, channel_path)
            try:
                package_and_upload_output_data_dir(
                    local_path=local_channel_path,
                    storage_dir_path=channel['value'],
                    scheme='gs'
                )
            except AssertionError as exception:
                logger.info(
                     "Skipping channel. {} is not a "
                     "directory or could not be found".format(
                         local_channel_path
                    )
                )

    def setup_model_artifacts(self):
        """ Iterates and downloads assets remoate -> model channel
        """
        logger.info("Setup model assets in {}".format(self.model_dir))
        # only fetch channels of environment prefix MLDOCK_MODEL_INPUT_CHANNEL_
        channels = self.get_model_input_channel_iter()

        if len(channels) == 0:
            logger.info("No input channels were found in ENV VARS.")

        for channel in channels:
            try:
                download_input_assets(
                    storage_dir_path=channel['value'],
                    local_path=self.model_dir,
                    scheme='gs'
                )

            except FileExistsError as exception:
                logger.info("{} Channel skipped. Already exists".format(channel['key']))

    def cleanup_model_artifacts(self):
        """ Iterates and uploads from model channel -> remote
        """
        logger.info("Cleanup model assets in {}".format(self.model_dir))

        # only fetch channels of environment prefix MLDOCK_MODEL_OUTPUT_CHANNEL_
        channels = self.get_model_output_channel_iter()

        if len(channels) == 0:
            logger.info("No model channels were found in ENV VARS.")

        for channel in channels:
            channel_path = channel['key'].replace("MLDOCK_MODEL_OUTPUT_CHANNEL_", "").lower()
            local_channel_path = Path(self.model_dir, channel_path)

            try:
                package_and_upload_model_dir(
                    local_path=local_channel_path,
                    storage_dir_path=channel['value'],
                    scheme='gs'
                )
            except AssertionError as exception:
                logger.info(
                     "Skipping channel. {} is not a "
                     "directory or could not be found".format(
                         local_channel_path
                    )
                )



"""Test dockerhub authentication"""
from mldock.platform_helpers.dockerhub import auth
from mldock.platform_helpers.utils import set_env

class TestDockerhubAuth:

    def test_get_dockerhub_credentials_success(self):

        env_vars = {
            "DOCKERHUB_USERNAME": "USERNAME",
            "DOCKERHUB_PASSWORD": "PASSWORD",
            "DOCKERHUB_REGISTRY": "https://index.docker.io/v1/",
            "DOCKERHUB_REPO": "REPONAME"
        }
        valid_vars = [{"key": key, 'value': value} for key,value in env_vars.items()]

        with set_env(**env_vars):

            username, password, registry, repository = auth.get_dockerhub_credentials()

            assert username == 'USERNAME'

            assert password == 'PASSWORD'

            assert registry == 'https://index.docker.io/v1/'

            assert repository == 'REPONAME'




"""Container API methods"""

import os
import sys
from pathlib import Path
import logging
from github import Github
from environs import Env
from mldock.platform_helpers import utils

logger = logging.getLogger('mldock')
env = Env()
env.read_env()  # read .env file, if it exists


def init_from_template(template_name, templates_root, src_directory, container_only: bool = False, template_server='local'):

    logger.info("Initializing Container Project")

    if template_server == 'local':
        template_dir = templates_root

    elif template_server == 'github':

        # set up in either ENV or configure local
        github_token = env("MLDOCK_GITHUB_TOKEN", default=None)
        github_org = env("MLDOCK_GITHUB_ORG", default=None)
        github_repo = env("MLDOCK_GITHUB_REPO", default=None)
        github_branch = env("MLDOCK_GITHUB_REPO_BRANCH", default=None)

        if github_token is None:
            raise KeyError(
                "Template server == 'github' requires the "
                "following environment variables: "
                "MLDOCK_GITHUB_TOKEN, MLDOCK_GITHUB_ORG, "
                "MLDOCK_GITHUB_REPO, MLDOCK_GITHUB_REPO_BRANCH"
            )

        github = Github(github_token)
        template_dir = utils.download_from_git(
            github,
            template_name,
            github_org,
            github_repo,
            github_branch,
            root=templates_root
        )

    logger.debug("Initializing template based on = {}".format(template_name))
    logger.debug("Template directory = {}".format(template_dir))

    logger.info("Setting up workspace") 

    if container_only:
        src_container_directory = os.path.join(
            src_directory,
            'container'
        )
        utils._copy_boilerplate_to_dst(
            os.path.join(template_dir, template_name, 'src/container'),
            src_container_directory,
            remove_first=True
        )
    else:
        utils._copy_boilerplate_to_dst(
            os.path.join(template_dir, template_name, 'src/'),
            src_directory
        )



#!/usr/bin/env python
from __future__ import print_function, absolute_import
import sys;sys.path.insert(1, ".")  # Do not remove this
import subprocess
import sys
import logging
import os
import json
import sys
import traceback
from src.trainer import run_training
from src.container.assets import TrainingContainer, environment, logger

def train_function():
    """
    The function to execute the training.

    :param input_data_path: [str], input directory path where all the training file(s) reside in
    :param model_save_path: [str], directory path to save your model(s)
    :param hyperparams_path: [optional[str], default=None], input path to hyperparams json file.

    """

    logger.debug("\n\n --- Running Training Script ---\n\n")
    try:

        # custom recommender training code
        run_training()

        logger.debug('Training complete.')
    except Exception as e:
        # Write out an error file. This will be returned as the failureReason in the
        # DescribeTrainingJob result.
        trc = traceback.format_exc()
        with open(os.path.join('/opt/ml/output/failure'), 'w') as s:
            s.write('Exception during training: ' + str(e) + '\n' + trc)
        # Printing this causes the exception to be in the training job logs, as well.
        logger.error('Exception during training: ' + str(e) + '\n' + trc)
        raise

if __name__ == '__main__':

    training = TrainingContainer(environment=environment)
    try:
        training.startup()

        train_function()
    except Exception as exception:
        logger.error(exception)
    finally:
        training.cleanup()



"""
    Core config managers

    Config managers that are reused across mldock cli tool.
"""
import sys
import os
import json
import click
import logging
from pathlib import Path

from mldock.platform_helpers import utils


logger=logging.getLogger('mldock')

class WorkingDirectoryManager:
    """Base config manager with basic read, write and update functionality.
    """
    def __init__(self, base_dir):
        self.base_dir = base_dir
        prompt_input = True

        if not self.file_exists(self.base_dir):
            prompt_input = click.prompt("File not found. Create?", default="yes")

            if prompt_input:
                self.base_dir.mkdir(parents=True, exist_ok=True)

        self.make_asset_dirs()

    def get_working_dir(self):
        return self.base_dir

    def make_asset_dirs(self):
        """Create the directory structure, if not exists
        """
        logger.debug("Creating assets directories in working dir {} .".format(self.base_dir))

        try:
            self.input_data_dir.mkdir(parents=False, exist_ok=True)
            self.input_config_dir.mkdir(parents=False, exist_ok=True)
            self.model_dir.mkdir(parents=False, exist_ok=True)
            self.output_data_dir.mkdir(parents=False, exist_ok=True)

        except Exception as exception:
            logger.error(exception)

    @staticmethod
    def file_exists(filename: str) -> bool:
        """Check if file exists

        Args:
            filename (str): path to file

        Returns:
            bool: whether file exists
        """
        return os.path.exists(filename)

    @property
    def input_data_dir(self):
        return Path(self.base_dir, 'data')

    @property
    def input_config_dir(self):
        return Path(self.base_dir, 'config')

    @property
    def model_dir(self):
        return Path(self.base_dir, 'model')
    
    @property
    def output_data_dir(self):
        return Path(self.base_dir, 'output')

class BaseConfigManager:
    """Base config manager with basic read, write and update functionality.
    """
    def __init__(self, filepath: str, create: bool = False):
        self.filepath = filepath

        self.config = self.load_config(self.filepath, create=create)

    @staticmethod
    def touch(path: str):
        """create a blank json file, seeded with {}

        Args:
            path (str): path to file
        """
        logger.info(path)
        with open(path, 'a') as file_:
            json.dump({}, file_)

    @staticmethod
    def file_exists(filename: str) -> bool:
        """Check if file exists

        Args:
            filename (str): path to file

        Returns:
            bool: whether file exists
        """
        return os.path.exists(filename)

    def check_if_exists_else_create(self, file_name: str, create: bool):
        """check that mldock fileexists"""
        if not self.file_exists(file_name):
            if create:
                # deal with possiblity of nested directory
                utils._mkdir(Path(file_name).parents[0])
                # create file
                self.touch(file_name)
            else:
                logger.error("No MLDOCK container project found with dir = '{}/'. Confirm 'yes' to create.".format(Path(file_name).parents[0]))
                sys.exit(1)

    def load_config(self, file_name: str, create: bool) -> dict:
        """loads config from file

        Args:
            filename (str): path to config to load
        Returns:
            dict: config
        """        
        self.check_if_exists_else_create(
            file_name=file_name, create=create
        )

        with open(file_name, "r") as file_:
            config = json.load(file_)
            return config

    def pretty_print(self):
        """pretty prints a json config to terminal
        """
        pretty_config = json.dumps(self.config, indent=4, separators=(',', ': '), sort_keys=True)
        logger.debug("{}\n".format(pretty_config))

    def write_file(self):
        """
        Trains ML model(s) locally
        :param dir: [str], source root directory
        :param docker_tag: [str], the Docker tag for the image
        :param image_name: [str], The name of the Docker image
        """
        logger.info(self.filepath)
        with open(self.filepath, 'w') as config_file:
            json.dump(self.config, config_file, indent=4, sort_keys=True)

    def get_config(self) -> dict:
        """get config object

        Returns:
            dict: config
        """
        return self.config




class Client:
    """docker client mock"""
    @staticmethod
    def login(
        username: str,
        password: str,
        email: str,
        registry: str,
        reauth: bool,
        **kwargs
    ):
        return {
            'username':'USERNAME',
            'password': 'PASSWORD',
            'registry': 'https://index.docker.io/v1/',
            'repository': 'REPONAME'
        }

def from_env():

    return Client()


from . import predictor as myapp

# This is just a simple wrapper for gunicorn to find your app.
# If you want to change the algorithm file, simply change "predictor" above to the
# new file.

app = myapp.app



"""Test GS storage helpers"""
from pathlib import Path
import tempfile
from mock import patch
from mldock.platform_helpers.gcp import storage
from tests.mocks.google.storage import GSBucket, GSBucketObject

class TestGSStorage:
    """Test google storage helpers"""

    @staticmethod
    def setup_storage_mocks(bucket_name, filename, storage_destination, MockStorageClient):
        """
            setup up storage mocks for GS storage

            returns:
                (bucket: GSBucket, blob: GSBucketObject)
        """
        bucket = GSBucket(name=bucket_name, objects=[filename])
        blob = GSBucketObject(name=storage_destination, bucket=bucket)
        MockStorageClient().get_bucket.return_value = bucket
        return bucket, blob

    @patch('mldock.platform_helpers.gcp.storage.StorageClient')
    @patch('mldock.platform_helpers.gcp.storage.download_blob_to_filename')
    def test_download_blob_to_filename_recieved_correct_args(self, mock_download_blob_to_filename, MockStorageClient):
        """
            test that download_blob_to_filename recieves the correct arguments
            during download_folder.
        """
        bucket_name = 'some-bucket'
        prefix_path = 'some/prefix'
        filename = 'file.txt'
        local_path = 'tmp/local/path'

        storage_destination = Path(prefix_path, filename).as_posix()

        bucket, blob = self.setup_storage_mocks(
            bucket_name,
            filename,
            storage_destination,
            MockStorageClient
        )

        storage.download_folder(
            bucket_name=bucket_name,
            prefix=prefix_path,
            target=local_path
        )

        args, kwargs = mock_download_blob_to_filename.call_args

        assert args == (blob, prefix_path, filename, local_path), (
            "Failed. download_blob_to_filename recieved the wrong arguments"
        )

    @patch('mldock.platform_helpers.gcp.storage._iter_nested_dir')
    @patch('mldock.platform_helpers.gcp.storage.StorageClient')
    @patch('mldock.platform_helpers.gcp.storage.upload_blob_from_filename')
    def test_upload_blob_from_filename_recieved_correct_args(
        self,
        mock_upload_blob_from_filename,
        MockStorageClient,
        mock__iter_nested_dir
    ):
        """
            test that upload_blob_from_filename recieves the correct arguments
            during upload_folder.
        """
        bucket_name = 'some-bucket'
        prefix_path = 'some/prefix'
        filename = 'file.txt'

        storage_destination = Path(prefix_path, filename).as_posix()

        with tempfile.TemporaryDirectory() as tmp_dir:
            local_path = tmp_dir
            path = Path(local_path, filename)
            path.touch()

            bucket, _ = self.setup_storage_mocks(
                bucket_name,
                filename,
                storage_destination,
                MockStorageClient
            )
            mock__iter_nested_dir.return_value = [path]

            storage.upload_folder(
                local_path=local_path,
                bucket_name=bucket_name,
                prefix=prefix_path
            )
            args, kwargs = mock_upload_blob_from_filename.call_args

        assert args == (bucket, path, storage_destination), (
            "Failed. upload_blob_from_filename recieved the wrong arguments"
        )



"""
    Container Setup Utilities

    Implements workflows based on the environment and GCP specific code to
    download and upload ML assets e.g. Input data, Models and output data
"""
import logging

logger = logging.getLogger('mldock')

class BaseTrainingContainer:
    """
        A set of tasks for setup and cleanup of container
    """
    def __init__(self, environment):
        self.environment = environment

    def startup(self):
        logger.info("\n\n --- Running Startup Script ---")
        if self.environment.environment_variables('MLDOCK_STAGE', default=None) == "prod":
            logger.info("Env == Prod")
            self.environment.setup_inputs()
            self.environment.setup_model_artifacts()
        logger.info("--- Setup Complete --- \n\n")

    def cleanup(self):
        """clean up tasks executed on container task complete"""
        logger.info("\n\n --- Running Cleanup Script ---")
        if self.environment.environment_variables('MLDOCK_STAGE', default=None) == "prod":
            logger.info("Env == Prod")
            self.environment.cleanup_model_artifacts()
            self.environment.cleanup_outputs()
        logger.info("--- Cleanup Complete --- \n\n")

class BaseServingContainer:
    """
        A set of tasks for setup and cleanup of container
    
        note:
            - Only supports a startup script. Cleanup is a bit fuzzy for serving.
    """
    def __init__(self, environment):
        self.environment = environment

    def startup(self):
        logger.info("\n\n --- Running Startup Script ---")
        
        if self.environment.environment_variables('MLDOCK_STAGE', default=None) == "prod":
            logger.info("Env == Prod")
            self.environment.setup_model_artifacts()
        logger.info("\n\n --- Setup Complete --- \n\n")

    def cleanup(self):
        """clean up tasks executed on container task complete"""
        pass



from pathlib import Path
import glob
import setuptools
from mldock.__version__ import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

def get_package_data(relative_path, relative_to):

    paths = Path(relative_to,relative_path).glob('**/*')
    data_files = [p.relative_to(relative_to).as_posix() for p in paths]

    return data_files

setuptools.setup(
    name="mldock",
    version=__version__,
    author="SheldonGrant",
    author_email="sheldz.shakes.williams@gmail.com",
    description="A docker tool that helps put machine learning in places that empower ml developers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SheldonGrant/mldock",
    packages=setuptools.find_packages(where='.'),
    package_data={
        'mldock': get_package_data(
            relative_path='templates',
            relative_to='mldock'
        )
    },
    setup_requires=['setuptools>=39.1.0'],
    install_requires=['future', 'environs', 'PyGithub'],
    extras_require={
        'gcp': ['google-cloud-storage', 'google-api-python-client'],
        'aws': ['boto3'],
        'cli': ['click', 'docker', 'future', 'requests', 'boto3', 'google-auth', 'appdirs', 'halo', 'PyYAML'],
        'sagemaker': ['sagemaker-training'],
        'testing': ['pandas', 'numpy', 'scipy', 'protobuf>=3.1']
    },
    entry_points="""
        [console_scripts]
        mldock=mldock.__main__:cli
    """,
    keywords=["docker", "machine learning", "ml", "ml services", "MLaaS"],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6"
    ],
    python_requires='>=3.6',
)



import sys
import os
import contextlib
import json
from urllib.parse import urlparse
from pathlib import Path
import tarfile
import zipfile
from typing import Iterator
import logging
import re
from distutils.dir_util import copy_tree, remove_tree
from distutils.dir_util import mkpath
from distutils.file_util import write_file
import tempfile
import base64

from github import GithubException
from github.GithubException import UnknownObjectException

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('mldock')

def _mkdir(dir_path: str):
    """make directory structure

    Args:
        dir_path (str): directory path to build
    """
    p = Path(dir_path)
    p.mkdir(parents=True, exist_ok=True)

def _iter_nested_dir(root_dir: str) -> Iterator[str]:
    """Iterate through nested folders.

    Args:
        root_dir (str): name of the directory to start form

    Yields:
        Iterator[str]: path iterator
    """
    rootdir = Path(root_dir)
    for path in rootdir.glob('**/*'):
        yield path

def _delete_file(file_path):
    """Delete a file

    Args:
        file_path (str): path to file

    Raises:
        TypeError: Expected a file_path of type file. Cannot be directory or other.
    """
    path = Path(file_path)
    if not path.is_file():
        raise TypeError("Expected a file_path of type file. Only deletes files")

    os.remove(path)

def _check_if_cloud_scheme(url: str, scheme: str = 's3') -> bool:
    parsed_url = urlparse(url)
    return parsed_url.scheme == scheme

def parse_url(url: str, scheme: str = 's3'):
    """Returns an (s3 bucket, key name/prefix) tuple from a url with an s3
    scheme.
    Args:
        url (str):
    Returns:
        tuple: A tuple containing:
            - str: S3 bucket name
            - str: S3 key
    """
    parsed_url = urlparse(url)
    if parsed_url.scheme != scheme:
        raise ValueError("Expecting '{}' scheme, got: {} in {}.".format(scheme, parsed_url.scheme, url))
    return parsed_url.netloc, parsed_url.path.lstrip("/")

def zip_folder(dir_path, output_file, rm_original=True):
    """zip in directory and optionally throw away unzipped"""

    with zipfile.ZipFile(output_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
        for file in _iter_nested_dir(dir_path):
            if file.name != Path(output_file).name and file.is_file():
                zipf.write(os.path.join(file), arcname=file.relative_to(dir_path))
                if rm_original:
                    _delete_file(file)

def zip_folder_as_tarfile(dir_path, output_file, rm_original=True):
    """zip folder as tarfile and optionally throw away original files"""
    with tarfile.open(output_file, "w:gz") as tar:
        for file in _iter_nested_dir(dir_path):
            if file.name != Path(output_file).name and file.is_file():
                tar.add(file, arcname=file.relative_to(dir_path))
                if rm_original:
                    _delete_file(file)

def unzip_file(filename, output_dir, rm_zipped=True):
    """unzip in directory and optionally throw away zipped"""
    with zipfile.ZipFile(filename, 'r', zipfile.ZIP_DEFLATED) as zipf:
        logger.info("Unzipping {} => {}".format(filename, output_dir))
        zipf.extractall(output_dir)
        if rm_zipped:
            logger.info("Removing {}".format(filename))
            _delete_file(filename)

def unzip_file_from_tarfile(filename, output_dir, rm_zipped=True):
    """untar in directory and optionally throw away zipped"""
    with tarfile.open(filename, "r:gz") as tar:
        logger.info("Unzipping {} => {}".format(filename, output_dir))
        tar.extractall(output_dir)
        if rm_zipped:
            logger.info("Removing {}".format(filename))
            _delete_file(filename)

def strip_scheme(url):
    parsed = urlparse(url)
    scheme = "%s://" % parsed.scheme
    return parsed.geturl().replace(scheme, '', 1)

def _read_json(path):
    """Read a JSON file.
    Args:
        path (str): Path to the file.
    Returns:
        (dict[object, object]): A dictionary representation of the JSON file.
    """
    with open(path, "r") as f:
        return json.load(f)

def _write_json(obj, path):
    """Write a serializeable object as a JSON file."""
    with open(path, "w") as f:
        json.dump(obj, f, indent=4)
        f.write('\n')

def _write_file(filepath, parents=True):
    """write a file"""
    if parents == True:
        mkpath(str(Path(filepath).parents[0].absolute()))
    write_file(str(Path(filepath).absolute()), '')

def _rename_file(base_path, current_filename, new_filename):
    """renames filename for a given base_path, saving the file in the same base_path

    Args:
        base_path ([type]): directory path containing file to rename
        current_filename ([type]): current name of the file to rename
        new_filename ([type]): new name for the renamed file
    """
    Path(base_path, current_filename).rename(Path(base_path, new_filename))

def _create_empty_file(base_path, filename):
    """renames filename for a given base_path, saving the file in the same base_path

    Args:
        base_path ([type]): directory path containing file to rename
        current_filename ([type]): current name of the file to rename
        new_filename ([type]): new name for the renamed file
    """
    Path(base_path, filename).touch(exist_ok=True)

def _copy_boilerplate_to_dst(src: str, dst: str, remove_first=False):
    """[summary]

    Args:
        src (str): [description]
        dst (str): [description]
    """
    source_path = str(Path(src).absolute())
    destination_path = str(Path(dst).absolute())
    if remove_first == True:
        try:
            logger.debug("removing first")
            remove_tree(destination_path)
        except FileNotFoundError as exception:
            logger.debug("No file found. Assuming already deleted.")
    response = copy_tree(source_path, destination_path)

@contextlib.contextmanager
def set_env(**environ):
    """
    Temporarily set the process environment variables.

    >>> with set_env(PLUGINS_DIR=u'test/plugins'):
    ...   "PLUGINS_DIR" in os.environ
    True

    >>> "PLUGINS_DIR" in os.environ
    False

    :type environ: dict[str, unicode]
    :param environ: Environment variables to set
    """
    old_environ = dict(os.environ)
    os.environ.update(environ)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)

def get_env_vars(environment, regex, flat=True):
    """Get all environ vars matching contains='<PREFIX>'"""
    if flat:
        filtered_env_vars = []
    else:
        filtered_env_vars = {}

    regex_pattern = re.compile(regex)

    for key,value in environment.items():
        result = regex_pattern.match(key)
        if result:
            # get all keys matching the regex
            if flat:
                filtered_env_vars.append({'key': key, 'value': value})
            else:
                filtered_env_vars.update({key: value})
    return filtered_env_vars

def get_sdk_credentials_volume_mount(
    auth_type='gcloud',
    **kwargs
):
    """
        configures a volume mount for providing provider resource
        permissions.
    """

    if auth_type == 'gcloud':
        host_path = kwargs.get('host_path', '~/.config/gcloud')
        container_path = kwargs.get('container_path', '/root/.config/gcloud')
    elif auth_type == 'awscli':
        host_path = kwargs.get('host_path', '~/.aws')
        container_path = kwargs.get('container_path', '/root/.aws')
    else:
        raise Exception("AUTH_TYPE = {} is not currently supported.")

    host_volume = Path(os.path.expanduser(host_path)).absolute().as_posix()
    container_bind = {'bind': container_path, 'mode': 'rw'}
    return {host_volume: container_bind}

def get_sha_for_tag(repository, tag):
    """
    Returns a commit PyGithub object for the specified repository and tag.
    """
    branches = repository.get_branches()
    matched_branches = [match for match in branches if match.name == tag]
    if matched_branches:
        return matched_branches[0].commit.sha

    tags = repository.get_tags()
    matched_tags = [match for match in tags if match.name == tag]
    if not matched_tags:
        raise ValueError('No Tag or Branch exists with that name')
    return matched_tags[0].commit.sha

def download_directory(repository, sha, server_path, local_prefix, relative_to):
    """
    Download all contents at server_path with commit tag sha in
    the repository.
    """
    try:
        contents = repository.get_contents(server_path, ref=sha)

        for content in contents:
            logger.info("Downloading {}".format(content.path))
            local_path = Path(local_prefix, Path(content.path).relative_to(relative_to))
            if content.type == 'dir':
                local_path.mkdir(parents=True, exist_ok=True)
                download_directory(
                    repository=repository,
                    sha=sha,
                    server_path=content.path,
                    local_prefix=local_prefix,
                    relative_to=relative_to
                )
            else:
                file_content = repository.get_contents(content.path, ref=sha)
                file_data = base64.b64decode(file_content.content)

                # set executables with chmod +x
                # given that mldock containers always have the "executor.sh"
                if local_path.name == 'executor.sh':
                    local_path.touch(0o777)
                else:
                    local_path.touch()

                with open(local_path, "w+") as file_out:
                    file_out.write(file_data.decode())
                    file_out.write("\n")
    except (GithubException, IOError) as exc:

        # get exc info
        exc_type, exc_value, exc_traceback = sys.exc_info()

        # if Unknown Object, inform user path is missing.
        if exc_type == UnknownObjectException:
            new_error = (
                "Server path = '{SERVER_PATH}' was "
                "not found in repository = '{REPOSITORY_NAME}'.".format(
                    SERVER_PATH=server_path,
                    REPOSITORY_NAME=repository.name
                )
            )
            logger.error(new_error)
            sys.exit(-1)
        raise

def download_from_git(github, start_dir, org, repo, branch, root: str = '.'):
    """Download the server path from git remote directory"""
    organization = github.get_organization(org)
    repository = organization.get_repo(repo)
    sha = get_sha_for_tag(repository, branch)

    tmp_dir = tempfile.mkdtemp()
    ## download to tmp
    download_directory(
        repository,
        sha,
        server_path=Path(root, start_dir).as_posix(),
        local_prefix=tmp_dir,
        relative_to=root
    )

    return tmp_dir



"""
    Serving Endpoint
    This is the file that implements a flask server to do inferences. It's the file that you will
    modify to implement the scoring for your own algorithm.
"""
import json
import flask
import numpy as np
import src.prediction as model_serving
from mldock.platform_helpers.mldock.inference.content_decoders import \
    numpy as content_decoders
from mldock.platform_helpers.mldock.inference.content_encoders import \
    numpy as content_encoders

app = flask.Flask(__name__)

@app.route('/ping', methods=['GET'])
def ping():
    """Determine if the container is working and healthy"""
    return flask.Response(response='\n', status=200, mimetype='application/json')


@app.route('/invocations', methods=['POST'])
def transformation():
    """Do an inference on a single batch of data. In this sample server, we take data as JSON"""
    if flask.request.content_type == 'application/json':
        data = flask.request.get_json()

        results = model_serving.handler(
            np.array(data)
        )
        return flask.Response(
            response=content_encoders.array_to_json(results),
            status=200,
            mimetype='application/json'
        )
    elif flask.request.content_type == 'text/csv':
        data = flask.request.data

        results = model_serving.handler(
            content_decoders.csv_to_numpy(data),
        )

        return flask.Response(
            response=content_encoders.array_to_csv(results, quoted=False),
            status=200,
            mimetype='text/csv'
        )
    else:
        return flask.Response(
            response=json.dumps({'message': 'This predictor only supports JSON or CSV data'}),
            status=415,
            mimetype='application/json'
        )



"""
    AWS STORAGE HELPERS

    Provides a set of helpers to intereact with aws s3 storage api's for asset management tasks.
"""
import os
from pathlib import Path
import logging
import boto3
import os 

from mldock.platform_helpers.utils import \
    _delete_file, _mkdir, _iter_nested_dir, _check_if_cloud_scheme, \
        parse_url, zip_folder_as_tarfile, zip_folder, \
         unzip_file_from_tarfile, unzip_file

logger = logging.getLogger('mldock')

def download_file_from_bucket(bucket, blob, prefix, filename, target):

    bucket_name = blob.bucket_name
    fullpath = os.path.join(bucket_name, prefix, filename)
    dst_filepath = os.path.join(target, os.path.basename(prefix))
    file_destination = os.path.join(dst_filepath, filename)

    logger.info("Download {} to local at {}".format(filename, file_destination))
    # make directory
    _mkdir(dst_filepath)
    # download
    bucket.download_file(blob.key, file_destination)

def download_folder(
    bucket_name: str,
    prefix: str,
    target: str
):
    """download folder from google cloud storage, ignoring folders.

    Args:
        bucket_name (str): [description]
        prefix (str): [description]
        target (str): [description]
    """
    logger.info("Starting Folder download from s3 storage")
    s3_resource = boto3.resource('s3')
    bucket = s3_resource.Bucket(bucket_name)

    for blob in bucket.objects.filter(Prefix = prefix):
        filename = Path(blob.key).name

        if len(filename) > 0:
            download_file_from_bucket(bucket, blob, prefix, filename, target)

def upload_file_to_bucket(bucket, path, storage_destination):

    logger.info("Upload {} to cloud at {}".format(path, storage_destination))
    bucket.upload_file(path.as_posix(), storage_destination)

def upload_folder(local_path: str, bucket_name: str, prefix: str):
    """Upload folder and contents to cloud storage

    Args:
        local_path (str): path to local directory
        bucket_name (str): bucket name to upload files to
        prefix (str): cloud filepath to write files to
    """
    logger.info("Starting Folder Upload to cloud storage")
    # get bucket
    s3_resource = boto3.resource("s3")
    bucket = s3_resource.Bucket(bucket_name)

    assert os.path.isdir(local_path), "'{}' is not a directory".format(local_path)

    for path in _iter_nested_dir(local_path):
        filepath = path.relative_to(local_path)
        storage_destination = os.path.join(prefix, filepath)
        if path.is_file():
            upload_file_to_bucket(bucket, path, storage_destination)

def download_input_assets(storage_dir_path: str, local_path: str, scheme: str):
    """download input asset folder from path, given that path is cloud storage path.

    Args:
        storage_dir_path (str): path to data in cloud storage
        local_path (str): where data should be stored (this comes from the environment).
        scheme (str, optional): scheme for cloud storage url. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)

    if is_cloud_storage:
        # download data if cloud storage path
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        download_folder(bucket_name=bucket, prefix=prefix, target=local_path)

        for file_path in Path(local_path).glob('*/*'):
            if file_path.suffix == '.zip':
                unzip_file(filename=file_path, output_dir=file_path.parent, rm_zipped=True)
            elif file_path.suffix == '.gz':
                unzip_file_from_tarfile(filename=file_path, output_dir=file_path.parent, rm_zipped=True)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")

def package_and_upload_model_dir(local_path: str, storage_dir_path: str, scheme: str):
    """
        Packages model/ dir as .tar.gz and uploads to cloud storage, given that storage_dir_path
        is cloud storage url path.

        Args:
            local_path (str): path to model directory
            storage_dir_path (str): cloud storage path destination
            scheme (str, optional): cloud storage url scheme. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)
    if is_cloud_storage:
        zip_folder_as_tarfile(
            dir_path=local_path,
            output_file=os.path.join(local_path, 'model.tar.gz'),
            rm_original=True
        )
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        upload_folder(local_path=local_path, bucket_name=bucket, prefix=prefix)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")

def package_and_upload_output_data_dir(local_path: str, storage_dir_path: str, scheme: str):
    """
        Packages output/ dir as .tar.gz and uploads to cloud storage, given that storage_dir_path
        is cloud storage url path.

        Args:
            local_path (str): path to model directory
            storage_dir_path (str): cloud storage path destination
            scheme (str, optional): cloud storage url scheme. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)

    if is_cloud_storage:
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        upload_folder(local_path=local_path, bucket_name=bucket, prefix=prefix)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")



from mock import patch
from click.testing import CliRunner

import logging
import tempfile
from mldock.__main__ import cli
from mldock.platform_helpers import utils

logger=logging.getLogger('mldock')

class TestContainerCommands:

    @staticmethod
    @patch(
        'future.moves.subprocess.check_output',
        return_value=None
    )
    def test_command_init_w_no_prompt_runs_successfully(future_output_mock):
        """
            Tests that the init no-prompt command runs successfully.
            Does not test that files are copied correctly.
        """
        runner = CliRunner()

        with tempfile.TemporaryDirectory('my_app') as tmp_dir:

            utils._copy_boilerplate_to_dst(
                src='tests/command/fixtures/base_container',
                dst=tmp_dir,
                remove_first=True
            )

            result = runner.invoke(
                cli=cli,
                args=['configure', 'local'],
                input='2\n\n'
            )
            result = runner.invoke(cli=cli, args=['container', 'init', '--dir', tmp_dir, '--no-prompt'])

        assert result.exit_code == 0, result.output

    @staticmethod
    @patch(
        'future.moves.subprocess.check_output',
        return_value=None
    )
    def test_command_init_w_user_input_runs_successfully(future_output_mock):
        """
            Tests that the init with user input command runs successfully.
            Does not test that files are copied correctly.
        """
        runner = CliRunner()
        with tempfile.TemporaryDirectory('my_app') as tmp_dir:

            utils._copy_boilerplate_to_dst(
                src='tests/command/fixtures/base_container',
                dst=tmp_dir,
                remove_first=True
            )

            result = runner.invoke(
                cli=cli,
                args=['configure', 'local'],
                input='2\n\n'
            )
            result = runner.invoke(
                cli=cli,
                args=['container', 'init', '--dir', tmp_dir],
                input='my_app\ngeneric\nsrc\ncontainer\nrequirements.txt\n\n\n\n\n'
            )

        assert result.exit_code == 0, result.output



import os
import sys
import json
import logging
import click
from pathlib import Path

from mldock.platform_helpers import utils

click.disable_unicode_literals_warning = True
logger = logging.getLogger('mldock')
MLDOCK_CONFIG_NAME = 'mldock.json'


@click.group()
def templates():
    """
    Commands to create, update and manage container projects and templates.
    """
    pass

@click.command()
@click.option('--name', help='Container name, used to name directories,etc.', required=True, type=str)
@click.option('--dir', help='Relative name of MLDOCK project', required=True, type=str)
@click.option('--out', help='Destination of template should be stored once created.', required=True, type=str)
@click.pass_obj
def create(obj, name, dir, out):
    """
    Command to create a mldock enabled container template
    """

    try:
        if not Path(dir, MLDOCK_CONFIG_NAME).exists():
            raise Exception("Path '{}' was not an mldock project. Confirm this directory is correct, otherwise create one.".format(dir))

        mldock_src_path = Path(dir, 'src')

        destination_path = Path(out, name)
        destination_path.mkdir(parents=False, exist_ok=True)

        destination_src_path = Path(destination_path, 'src')

        utils._copy_boilerplate_to_dst(mldock_src_path, destination_src_path)
    except Exception as exception:
        logger.error(exception)
        raise

templates.add_command(create)



"""
    GCP STORAGE HELPERS

    Provides a set of helpers to intereact with gcp gs storage api's for asset management tasks.
"""
import os
from pathlib import Path
import logging
from google.cloud.storage import Client as StorageClient

from mldock.platform_helpers.utils import \
    _delete_file, _mkdir, _iter_nested_dir, _check_if_cloud_scheme, \
        parse_url, zip_folder_as_tarfile, zip_folder, \
         unzip_file_from_tarfile, unzip_file

logger = logging.getLogger('mldock')

def download_blob_to_filename(blob, prefix, filename, target):
    """download blob from google storage to filename"""
    bucket_name = blob.bucket.name
    fullpath = os.path.join(bucket_name, prefix, filename)
    dst_filepath = os.path.join(target, os.path.basename(prefix))
    file_destination = os.path.join(dst_filepath, filename)
    logger.info("Download {} to local at {}".format(filename, file_destination))
    # make directory
    _mkdir(dst_filepath)
    # download
    blob.download_to_filename(file_destination)

def download_folder(
    bucket_name: str,
    prefix: str,
    target: str
):
    """download folder from google cloud storage, ignoring folders.

    Args:
        bucket_name (str): [description]
        prefix (str): [description]
        target (str): [description]
    """
    logger.info("Starting Folder download from cloud storage")
    storage_client = StorageClient()
    bucket = storage_client.get_bucket(bucket_name)
    blobs = bucket.list_blobs(prefix=prefix)  # Get list of files
    for blob in blobs:
        filename = Path(blob.name).name
        if len(filename) > 0:
            download_blob_to_filename(blob, prefix, filename, target)

def upload_blob_from_filename(bucket, path, storage_destination):
    """upload blob to google storage from local filename"""
    logger.info("Upload {} to cloud at {}".format(path, storage_destination))
    blob = bucket.blob(storage_destination)
    blob.upload_from_filename(path)

def upload_folder(local_path: str, bucket_name: str, prefix: str):
    """Upload folder and contents to cloud storage

    Args:
        local_path (str): path to local directory
        bucket_name (str): bucket name to upload files to
        prefix (str): cloud filepath to write files to
    """
    logger.info("Starting Folder Upload to cloud storage")
    # get bucket
    storage_client = StorageClient()
    bucket = storage_client.get_bucket(bucket_name)
    assert os.path.isdir(local_path), "'{}' is not a directory".format(local_path)

    for path in _iter_nested_dir(local_path):
        filepath = path.relative_to(local_path)
        storage_destination = os.path.join(prefix, filepath)
        if path.is_file():
            upload_blob_from_filename(bucket, path, storage_destination)

def download_input_assets(storage_dir_path: str, local_path: str, scheme: str = 'gs'):
    """download input asset folder from path, given that path is cloud storage path.

    Args:
        storage_dir_path (str): path to data in cloud storage
        local_path (str): where data should be stored (this comes from the environment).
        scheme (str, optional): scheme for cloud storage url. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)

    if is_cloud_storage:
        # download data if cloud storage path
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        download_folder(bucket_name=bucket, prefix=prefix, target=local_path)

        for file_path in Path(local_path).glob('*/*'):
            if file_path.suffix == '.zip':
                unzip_file(filename=file_path, output_dir=file_path.parent, rm_zipped=True)
            elif file_path.suffix == '.gz':
                unzip_file_from_tarfile(filename=file_path, output_dir=file_path.parent, rm_zipped=True)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")

def package_and_upload_model_dir(local_path: str, storage_dir_path: str, scheme: str = 'gs'):
    """
        Packages model/ dir as .tar.gz and uploads to cloud storage, given that storage_dir_path
        is cloud storage url path.

        Args:
            local_path (str): path to model directory
            storage_dir_path (str): cloud storage path destination
            scheme (str, optional): cloud storage url scheme. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)
    if is_cloud_storage:
        zip_folder_as_tarfile(
            dir_path=local_path,
            output_file=os.path.join(local_path, 'model.tar.gz'),
            rm_original=True
        )
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        upload_folder(local_path=local_path, bucket_name=bucket, prefix=prefix)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")

def package_and_upload_output_data_dir(local_path: str, storage_dir_path: str, scheme: str = 'gs'):
    """
        Packages output/ dir as .tar.gz and uploads to cloud storage, given that storage_dir_path
        is cloud storage url path.

        Args:
            local_path (str): path to model directory
            storage_dir_path (str): cloud storage path destination
            scheme (str, optional): cloud storage url scheme. Defaults to 'gs'.
    """
    is_cloud_storage = _check_if_cloud_scheme(url=storage_dir_path, scheme=scheme)

    if is_cloud_storage:
        bucket, prefix = parse_url(url=storage_dir_path, scheme=scheme)
        upload_folder(local_path=local_path, bucket_name=bucket, prefix=prefix)
    else:
        Exception("No Cloud storage url was found. Must have gs:// schema")



"""
    NUMPY CONTENT DECODERS

    Handle Decoding of Content in to appropriate format from request as numpy array.

    e.g. 
        list of lists -> np.array.
"""
import csv
import io
import json
import numpy as np

from mldock.platform_helpers.mldock.inference import content_types

def npy_to_numpy(npy_array):
    """Convert an NPY array into numpy.
    Args:
        npy_array (npy array): NPY serialized array to be converted.
    Returns:
        (np.array): Converted numpy array.
    """
    stream = io.BytesIO(npy_array)
    return np.load(stream, allow_pickle=True)

def json_to_numpy(json_data):
    """
        Convert a JSON object to a numpy array.

        Args:
            string_like (str): JSON serialized data.
        Returns:
            (np.array): data as Numpy array.
    """
    data = json.loads(json_data)
    return np.array(data)

def csv_to_numpy(bytes_like):
    """
        Convert a CSV object to a numpy array.

        Args:
            bytes_like (str): bytes serialized CSV string.
        Returns:
            (np.array): data as Numpy array.
    """
    try:
        stream = io.StringIO(bytes_like.decode())
        reader = csv.reader(stream, delimiter=",", quotechar='"', doublequote=True, strict=True)
        data = np.array([row for row in reader]).squeeze()
        return data

    except ValueError as e:
        if dtype is not None:
            raise Exception(
                "Error while writing numpy array: {}. dtype is: {}".format(e, dtype)
            )
        raise
    except Exception as e:
        raise Exception("Error while decoding csv: {}".format(e))



"""
    Terminal Utils

    Incl. styling and formatting in terminal commands.
"""
import logging
import click
import typing as t
from gettext import ngettext
from halo import Halo

logger = logging.getLogger('mldock')

def pretty_build_logs(line: dict):

    if line is None:
        return None

    error = line.get('error', None)
    errorDetail = line.get('errorDetail', None)

    if error is not None:
        logger.error('{}\n{}'.format(error, errorDetail))
    
    stream = line.get('stream', '')

    return " ==> {MESSAGE}".format(MESSAGE=stream)

class ProgressLogger:
    """
        A terminal based progress context manager.

        starts the progress spinner, and passes the object back.
        finally raises success or failure if exception is raised.    
    """
    def __init__(self, **kwargs):
        self.group = kwargs.pop('group', None)
        if self.group is None:
            self.group = kwargs.get('text', None)

        self.on_success = kwargs.pop('on_success', None)
        self.on_failure = kwargs.pop('on_failure', None)
        
        if len(self.group) > 0:
            self.group = self.group + " "

        self.spinner = Halo(**kwargs)

    def __enter__(self):
        self.spinner.start()
        return self.spinner

    def __exit__(self, exc_type, exc_value, exc_tb):
        # all though this always runs on complete
        # we would like behaviour to register either success or failure
        # use exc_type == None to raise success
        if exc_type is not None:
            error_msg = "Error raised during {}".format(self.group)
            if self.on_failure is not None:
                error_msg = self.on_failure
            self.spinner.fail(error_msg)
            return False # reraise the exception

        success_msg = "{}Complete".format(self.group)
        if self.on_success is not None:
            success_msg = self.on_success
        self.spinner.succeed(success_msg)
        return True

class ChoiceWithNumbers(click.Choice):
    """Extends the click.Choice class to allow user to make choice by number"""
    def convert(
        self, value: t.Any, param: t.Optional["Parameter"], ctx: t.Optional["Context"]
    ) -> t.Any:
        """
            Match through normalization and case sensitivity
            first do token_normalize_func, then lowercase
            preserve original `value` to produce an accurate message in
            `self.fail`
        """
        normed_value = value
        normed_choices = {choice: choice for choice in self.choices}
        try:
            # try coherse to integer else
            normed_value = int(normed_value)
            normed_value = self.choices[normed_value - 1]

        except Exception as exception:
            pass

        if ctx is not None and ctx.token_normalize_func is not None:
            normed_value = ctx.token_normalize_func(value)
            normed_choices = {
                ctx.token_normalize_func(normed_choice): original
                for normed_choice, original in normed_choices.items()
            }

        if not self.case_sensitive:
            normed_value = normed_value.casefold()
            normed_choices = {
                normed_choice.casefold(): original
                for normed_choice, original in normed_choices.items()
            }

        if normed_value in normed_choices:
            return normed_choices[normed_value]

        choices_str = ", ".join(map(repr, self.choices))
        self.fail(
            ngettext(
                "{value!r} is not {choice}.",
                "{value!r} is not one of {choices}.",
                len(self.choices),
            ).format(value=value, choice=choices_str, choices=choices_str),
            param,
            ctx,
        )

def style_2_level_detail(major_detail, minor_detail):
    """Styling for two level detail"""
    format_minor_detail = click.style(
        "({})".format(minor_detail),
        fg='bright_green'
    )
    return "{} {}".format(
        major_detail,
        format_minor_detail
    )

def style_dropdown(group_name: str, options: list, default: str = None):
    """Styling prompt with choice options as a drop down list"""
    options = options.copy()
    for i, option in enumerate(options):
        option_formatted = "[{id}] {option}".format(id=i+1, option=option)
        if option == default:
            option_formatted = click.style(option_formatted, fg='bright_blue')
        options[i] = option_formatted

    option_msg = "\n".join(options)
    return "(Select {group_name} from list)\n\n{option_msg}\n=>".format(
        group_name=click.style(group_name, fg='bright_blue'),
        option_msg=option_msg
    )
